#include <windows.h>
#include <stdio.h>
#include "getopt.h"
#include "winuser.h"

/* These should be changed if you make any code changes */
#define VERSION_MAJOR   0
#define VERSION_MINOR   2
#define VERSION_BUILD   0x4

/*
 * Command line option token values.
 *  0x0000-0x00ff   Single characters
 *  0x1000-0x1fff   Long switches (no arg)
 *  0x2000-0x2fff   Long settings (arg required)
 */
#define OPT_H           0x0001

#define OPT_HELP        0x1000
#define OPT_SAVE        0x1001
#define OPT_LIST        0x1002

#define OPT_X           0x2000
#define OPT_Y           0x2001
#define OPT_BPP         0x2002
#define OPT_VSYNC       0x2003

struct option long_options[] = {
    { "?", no_argument, NULL, OPT_H },
    { "h", no_argument, NULL, OPT_H },
    { "help", no_argument, NULL, OPT_HELP },
    { "save", no_argument, NULL, OPT_SAVE },
    { "list", no_argument, NULL, OPT_LIST },
    // Logical Volume Descriptor
    { "x", required_argument, NULL, OPT_X },
    { "y", required_argument, NULL, OPT_Y },
    { "vsync", required_argument, NULL, OPT_VSYNC },
    { "bp", required_argument, NULL, OPT_BPP },

    { NULL, no_argument, NULL, NULL }
};

ULONG opt_x = -1;
ULONG opt_y = -1;
ULONG opt_vsync = -1;
ULONG opt_bpp = -1;

BOOLEAN opt_save = FALSE;
BOOLEAN opt_list = FALSE;

#define CUR_DISPLAY_NAME     "Display1"
#define CUR_DISPLAY_DEV_NAME "\\\\.\\" CUR_DISPLAY_NAME

void
usage(void)
{
    fprintf(stderr, "Display Mode Control utility, Copyright (C) by AlterWare, 2003\n"
        "Version %d.%d (%x)\n", VERSION_MAJOR,
        VERSION_MINOR, VERSION_BUILD);
    fprintf(stderr,
        "Usage:\n"
        "\tvmctl <options> \n"
        "Options:\n"
        "\t--help, --h, --?      display this screen\n"
        "\t--x=NUMBER            X-resolution\n"
        "\t--y=NUMBER            Y-resolution\n"
        "\t--vsync=NUMBER        Vertical Sync. (Hz)\n"
        "\t--bp=NUMBER           bits per pixel\n"
        "\t--save                save new settings in registry (system-wide change)\n"
        "\t--list                display all available modes\n"
        "\n"

        "Example:\n"
        "\tvmctl --x=320 --y=240\n"
        "\tvmctl --x=1600 --y=1280 --ref=85\n"
    );
}

void
parse_args(int argc, char *argv[])
{
    int retval;
    while (argc > 1) {
            retval = getopt_long(argc, argv, "V", long_options, NULL);
        switch (retval)
        {
            case OPT_HELP:
            case OPT_H:
                usage();
                exit(0);
                break;

            case OPT_X:
                opt_x = strtoul(optarg, 0, 0);
                break;

            case OPT_Y:
                opt_y = strtoul(optarg, 0, 0);
                break;

            case OPT_VSYNC:
                opt_vsync = strtoul(optarg, 0, 0);
                break;

            case OPT_BPP:
                opt_bpp = strtoul(optarg, 0, 0);
                break;

            case OPT_SAVE:
                opt_save = TRUE;
                break;

            case OPT_LIST:
                opt_list = TRUE;
                break;

            case EOF:
                return;

            default:
                fprintf(stderr, "Try `vmctl --help' "
                    "for more information\n");
                exit(-2);
        }
    }

}

void main (int argc, char* argv[]) {

    DEVMODE DisplayMode;
    ULONG flags = 0;
    ULONG i=0;
//    ULONG vsync_list[128];
//    ULONG old_vsync_list[128];
//    ULONG vsync_list_sz;
//    ULONG old_vsync_list_sz;
//    ULONG last_i=0;
    
    ULONG old_x = -1;
    ULONG old_y = -1;
//    ULONG old_vsync = -1;
    ULONG old_bpp = -1;
    BOOLEAN printed = FALSE;
//    BOOLEAN scan = TRUE;

    if ( argc < 2 ) {
        usage();
        exit(-1);
    }
    parse_args(argc, argv);

    if(opt_list) {
        while(EnumDisplaySettings(CUR_DISPLAY_DEV_NAME, i, &DisplayMode)) {
            i++;
            if(opt_x != -1 &&
               DisplayMode.dmPelsWidth != opt_x)
                continue;
            if(opt_y != -1 &&
               DisplayMode.dmPelsHeight != opt_y)
                continue;
            if(opt_vsync != -1 &&
               DisplayMode.dmDisplayFrequency != opt_vsync)
                continue;
            if(opt_bpp != -1 &&
               DisplayMode.dmBitsPerPel != opt_bpp)
                continue;
            if(DisplayMode.dmPelsWidth  == old_x &&
               DisplayMode.dmPelsHeight == old_y &&
               DisplayMode.dmBitsPerPel == old_bpp) {
//                if(!scan)
                    printf("/%d", DisplayMode.dmDisplayFrequency);
//                vsync_list_sz++;
            } else {
                if(printed)
                    printf("Hz\n");
//                vsync_list_sz = 0;
                printf("%dx%d %dbit %d",
                       DisplayMode.dmPelsWidth,
                       DisplayMode.dmPelsHeight,
                       DisplayMode.dmBitsPerPel,
                       DisplayMode.dmDisplayFrequency
                       );
                printed = TRUE;
                old_x   = DisplayMode.dmPelsWidth;
                old_y   = DisplayMode.dmPelsHeight;
                old_bpp = DisplayMode.dmBitsPerPel;
            }
//            vsync_list[vsync_list_sz] = DisplayMode.dmDisplayFrequency;
        }
        if(printed)
            printf("Hz\n");
        exit(1);
    }

    EnumDisplaySettings(CUR_DISPLAY_DEV_NAME, ENUM_CURRENT_SETTINGS, &DisplayMode);

    if(opt_x != -1) {
        DisplayMode.dmPelsWidth = opt_x;
        flags |= DM_PELSWIDTH;
    }
    if(opt_y != -1) {
        DisplayMode.dmPelsHeight = opt_y;
        flags |= DM_PELSHEIGHT;
    }
    if(opt_vsync != -1) {
        DisplayMode.dmDisplayFrequency = opt_vsync;
        flags |= DM_DISPLAYFREQUENCY;
    }
    if(opt_bpp != -1) {
        DisplayMode.dmBitsPerPel = opt_bpp;
        flags |= DM_BITSPERPEL;
    }

    DisplayMode.dmFields = DM_DISPLAYFLAGS | flags;

    ChangeDisplaySettingsEx(CUR_DISPLAY_DEV_NAME,  &DisplayMode, NULL, 
        CDS_SET_PRIMARY, NULL);
    if(opt_save) {
        ChangeDisplaySettingsEx(CUR_DISPLAY_DEV_NAME,  &DisplayMode, NULL, 
            CDS_UPDATEREGISTRY, NULL);
    }
    exit(1);
}
